'''
    Convert an image file to a data uri.

    Copyright 2012-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import base64
import os.path

try:
    from django.conf import settings
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.python.log import get_log

log = get_log()
debugging = False

img_cache = {}

def data_image(filename, browser=None, mime_type=None):
    ''' Encode a file in base 64 as a data uri.

        The filename is relative to settings.STATIC_ROOT.
        'browser' is defined in denova.net.browser.

        If the data uri is too large for ie8 or anything goes wrong,
        returns the filename.

        >>> CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
        >>> filename = os.path.join(CURRENT_DIR, 'static', 'images', 'Kebab_icon.png')
        >>> data = data_image(filename)
        >>> log(f'data: {data}')
        >>> data.startswith('data:image/png;base64')
        True
        >>> filename = os.path.join(CURRENT_DIR, 'static', 'images', 'Kebab_icon.png')
        >>> imagename = data_image(filename, browser='opera')
        >>> filename == imagename
        True
    '''

    if browser:
        log_if_debugging(f'browser: {browser}')

    # data uris don't work well with all browsers
    # opera only supports images up to 4K so we just ignore them
    if browser and (
                    'unknown' in browser or
                    'ie5' in browser or
                    'ie6' in browser or
                    'ie7' in browser or
                    'opera' in browser or
                    'java' in browser):

        log_if_debugging(f'browser "{browser}" does not support data uri')
        if filename.startswith('/'):
            result = filename
        else:
            result = '/static/' + filename

    elif filename in img_cache:

        result = img_cache[filename]

    else:

        log_if_debugging(f'filename : {filename}')

        result = filename
        try:
            if not mime_type:
                basename = filename.split('/')[-1]
                data_type = basename.split('.')[-1].lower()
                if data_type == 'jpeg':
                    data_type = 'jpg'
                # until we handle other types
                if data_type not in ['png', 'jpg', 'gif']:
                    assert ValueError, "filename does not end in '.png', '.jpg', or '.gif', and no mime_type specified"
                mime_type = f'image/{data_type}'
                log_if_debugging(f'set mime_type to {mime_type}')

            # if there's no leading path, add one
            pathname = os.path.join(settings.STATIC_ROOT, filename)
            log_if_debugging(f'pathname = {pathname}')

            '''
                Only IE8 is officially to limited to 32K.
                Earlier versions of Internet Explorer don't support data uris,
                and later ones don't have this limit.
                See http://en.wikipedia.org/wiki/Data_URI_scheme.

                But other browsers (e.g. Firefox on linux) can bog down
                badly with large data uris.
            '''
            # the RFC 2397 says the max size is 2100, but most browsers
            #  that support data images support larger images
            max_size = 100 * 1024
            if browser:
                if ('ie8' in browser):
                    max_size = 32 * 1024
                elif ('ff5' in browser):
                    max_size = 20 * 1024

            log_if_debugging(f'os.path.getsize({pathname})')
            log_if_debugging(f'os.path.exists({pathname}) = {os.path.exists(pathname)}')
            if os.path.exists(pathname):
                log_if_debugging(f'os.path.getsize({pathname}) = {os.path.getsize(pathname)}')
                if os.path.getsize(pathname) < max_size:
                    log_if_debugging(f'file_to_data_uri({pathname}, {mime_type}')
                    result = file_to_data_uri(pathname, mime_type)
                    log_if_debugging(f'done file_to_data_uri({pathname}, {mime_type})')

                else:
                    #log_if_debugging('IE8 does not allow data uris larger than 32K{pathname})
                    log_if_debugging(f'No data uris larger than 32K: {pathname}')
                    result = pathname
            else:
                msg = f'missing datafile: {pathname}'
                log.warning(msg)
                result = f'<!-- WARNING: {msg} -->'

        except Exception as e:
            log.error(e)
            # like django template processing, fail quietly
            result = ''

        log_if_debugging(f'result: {result}')
        img_cache[filename] = result

    return result

def data_uri(data, mime_type, charset=None):
    '''
        Convert binary data to a data uri.

        >>> CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
        >>> filename = os.path.join(CURRENT_DIR, 'static', 'images', 'Kebab_icon.png')
        >>> f = open(filename, 'rb')
        >>> data = f.read()
        >>> uri = data_uri(data, 'image/png')
        >>> uri.startswith('data:image/png;base64')
        True
    '''

    data_string = base64.b64encode(data).decode('utf-8').replace('\n', '')
    if charset:
        uri = f'data:{charset};{mime_type};base64,{data_string}'
    else:
        uri = f'data:{mime_type};base64,{data_string}'

    log_if_debugging(f'uri: {uri}')

    return uri

def file_to_data_uri(filename, mime_type):
    '''
        Convert file to a data uri.

        >>> CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
        >>> filename = os.path.join(CURRENT_DIR, 'static', 'images', 'Kebab_icon.png')
        >>> datauri = file_to_data_uri(filename, 'image/png')
        >>> datauri.startswith('data:image/png;base64')
        True
    '''

    with open(filename, 'rb') as input_file:
        datauri = data_uri(input_file.read(), mime_type)

    return datauri

def log_if_debugging(message):
    ''' Log if debugging. '''

    if debugging:
        log.debug(message)



'''
    Save a python expression as a variable.

    Copyright 2019-2020 DeNova
    Last modified: 2020-10-22

    Django's philosophy is that code is for business logix and
    templates are for rendering. But sometimes calculations are
    needed purely for rendering, such as in calculating css
    opacity.

    If you don't calc in the template, you have to split rendering
    between python files and templates. That's a maintenance nightmare,
    much worse than expressions in a template.

    The expr tag can be misused to make Django templates much more like php,
    with no distinction between MVC model and view.

    From http://djangoclips.org/clips/9/:

        This tag can be used to calculate a python expression, and
        save it into a template variable which you can reuse later or
        directly output to template. So if the default django tag
        can not be suit for your need, you can use it.

        How to use it

            {% expr "1" as var1 %}
            {% expr [0, 1, 2] as var2 %}
            {% expr _('Menu') as var3 %}
            {% expr var1 + "abc" as var4 %}
            ...
            {{ var1 }}

        Syntax

            {% expr python_expression as variable_name %}

        python_expression can be valid python expression, and you can
        even use _() to translate a string. Expr tag also can use
        context variables.
'''

try:
    from django import template
    from django.utils.translation import gettext_lazy as _
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

import re

register = template.Library()

class ExprNode(template.Node):
    def __init__(self, expr_string, var_name):
        self.expr_string = expr_string
        self.var_name = var_name

    def render(self, context):
        clist = list(context)
        clist.reverse()
        d = {}
        d['_'] = _
        for c in clist:
            d.update(c)
        if self.var_name:
            context[self.var_name] = eval(self.expr_string, d)
            return ''
        else:
            return str(eval(self.expr_string, d))


r_expr = re.compile(r'(.*?)\s+as\s+(\w+)', re.DOTALL)


def do_expr(parser, token):
    try:
        tag_name, arg = token.contents.split(None, 1)
    except ValueError:
        raise template.TemplateSyntaxError(f"{token.contents[0]} tag requires arguments")
    m = r_expr.search(arg)
    if m:
        expr_string, var_name = m.groups()
    else:
        if not arg:
            raise template.TemplateSyntaxError(f"{tag_name} tag at least require one argument")

        expr_string, var_name = arg, None
    return ExprNode(expr_string, var_name)


do_expr = register.tag('expr', do_expr)



'''
    MIME constants.

    Copyright 2014-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

CONTENT_TYPE_KEYWORD = 'Content-Type'
CONTENT_XFER_ENCODING_KEYWORD = 'Content-Transfer-Encoding'
CONTENT_DISPOSITION_KEYWORD = 'Content-Disposition'

RETURN_PATH_KEYWORD = 'Return-Path'
DELIVERED_TO_KEYWORD = 'Delivered-To'
X_ORIGINAL_TO_KEYWORD = 'X-Original-To'
FROM_KEYWORD = 'From'
TO_KEYWORD = 'To'
CC_KEYWORD = 'Cc'
SUBJECT_KEYWORD = 'Subject'
DATE_KEYWORD = 'Date'


MESSAGE_ID_KEYWORD = 'Message-ID'
MIME_VERSION_KEYWORD = 'Mime-Version'
MIME_VERSION = '1.0'

PRIMARY_TYPE_DELIMITER = '/'
TEXT_PRIMARY_TYPE = 'text'
PLAIN_SUB_TYPE = 'plain'
TEXT_PLAIN_TYPE = f'{TEXT_PRIMARY_TYPE}{PRIMARY_TYPE_DELIMITER}{PLAIN_SUB_TYPE}'
HTML_SUB_TYPE = 'html'
TEXT_HTML_TYPE = f'{TEXT_PRIMARY_TYPE}{PRIMARY_TYPE_DELIMITER}{HTML_SUB_TYPE}'

MULTIPART_PRIMARY_TYPE = 'multipart'
ENCRYPTED_SUB_TYPE = 'encrypted'
ALTERNATIVE_SUB_TYPE = 'alternative'
MIXED_SUB_TYPE = 'mixed'
SIGNED_SUB_TYPE = 'signed'
MULTIPART_ENCRYPTED_TYPE = f'{MULTIPART_PRIMARY_TYPE}{PRIMARY_TYPE_DELIMITER}{ENCRYPTED_SUB_TYPE}'
MULTIPART_ALT_TYPE = f'{MULTIPART_PRIMARY_TYPE}{PRIMARY_TYPE_DELIMITER}{ALTERNATIVE_SUB_TYPE}'
MULTIPART_MIXED_TYPE = f'{MULTIPART_PRIMARY_TYPE}{PRIMARY_TYPE_DELIMITER}{MIXED_SUB_TYPE}'
MULTIPART_SIGNED_TYPE = f'{MULTIPART_PRIMARY_TYPE}{PRIMARY_TYPE_DELIMITER}{SIGNED_SUB_TYPE}'
APPLICATION_TYPE = 'application'
APPLICATION_ALT_TYPE = f'{APPLICATION_TYPE}{PRIMARY_TYPE_DELIMITER}{ALTERNATIVE_SUB_TYPE}'
PGP_SUB_TYPE = 'pgp-encrypted'
PGP_TYPE = f'{APPLICATION_TYPE}{PRIMARY_TYPE_DELIMITER}{PGP_SUB_TYPE}'
PGP_SIG_SUB_TYPE = 'pgp-signature'
PGP_SIG_TYPE = f'{APPLICATION_TYPE}{PRIMARY_TYPE_DELIMITER}{PGP_SIG_SUB_TYPE}'
OCTET_STREAM_SUB_TYPE = 'octet-stream'
OCTET_STREAM_TYPE = f'{APPLICATION_TYPE}{PRIMARY_TYPE_DELIMITER}{OCTET_STREAM_SUB_TYPE}'

PROTOCOL_KEYWORD = 'protocol'
CHARSET_KEYWORD = 'charset'
PGP_MIME_VERSION_FIELD = 'Version: 1'

MIME_7BIT = '7BITMIME'
MIME_8BIT = '8BITMIME'
BITS_7 = '7bit'
BITS_8 = '8bit'

QUOTED_PRINTABLE_ENCODING = 'quoted-printable'
BASE64_ENCODING = 'base64'




'''
    Redirect 404 response to another http server.

    The server is specified by settings.REDIRECT_404_SERVER.
    REDIRECT_404_SERVER is in the form:

        [USERNAME[:PASSWORD]@]HOST[:PORT]

    HOST can be a DNS name or IP address.
    REDIRECT_404_SERVER Examples::

        example.com
        example.com:8080
        username@example.com:8080
        username:password@example.com:8080
        127.0.0.1
        127.0.0.1:8080

    Copyright 2014-2020 DeNova
    Last modified: 2020-10-22

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

from traceback import format_exc
from urllib.parse import urlsplit, urlunsplit

try:
    from django import http
    from django.conf import settings
    from django.utils.deprecation import MiddlewareMixin
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.python.log import get_log

log = get_log()


class Redirect404Middleware(MiddlewareMixin):

    def process_response(self, request, response):

        try:
            if response.status_code == 404:

                log('got 404')
                url = request.get_full_path()
                parts = urlsplit(url)

                if parts.scheme == '':
                    if request.is_secure():
                        scheme = 'https'
                    else:
                        scheme = 'http'
                else:
                    scheme = parts.scheme
                new_host = settings.REDIRECT_404_SERVER

                new_url = urlunsplit([scheme, new_host,
                    parts.path, parts.query, parts.fragment])

                try:
                    response = http.HttpResponsePermanentRedirect(new_url)

                except Exception:
                    # just log it and return the exsting 404
                    log(f'url: {url}')
                    log('parts: %s' % repr(parts))
                    log(f'new url: {new_url}')
                    log(format_exc())

        except Exception:
            log(format_exc())
            raise

        return response



#!/usr/bin/env python3
'''
    Use openssl to generate a self-signed 4096 RSA cert.
    Requires denova.net.cli, which requires pexpect to be installed from pypi.

    To do: Sanitize denova.net.cli inputs.
    Do we need to include a "distinguished_name" or group or req name in the conf file?

    Copyright 2014-2020 DeNova
    Last modified: 2020-10-20
'''

import os
import re
from datetime import datetime, timedelta
from subprocess import CalledProcessError
from traceback import format_exc

from denova.os.cli import minimal_env, Responder
from denova.os.command import run
from denova.os.lock import locked
from denova.python.log import get_log

log = get_log()

PRIVATE_KEY = 'private.key'
PUBLIC_CERT = 'public.crt'

# if openssl  is looking for openssl.cnf, it's in /etc/ssl
# openssl wants the path in the env var OPENSSL_CNF
OPENSSL_CONF_FILENAME = '/etc/ssl/openssl.cnf'

SELF_SIGNED_CERT_ERR_MSG = 'self signed certificate'
EXPIRED_CERT_ERR_MSG = 'certificate expired on'

def verify_certificate(hostname, port, ca_certs_dir=None):
    '''
        Verify a certificate is valid. Compare against openssl's published certs.

        Debian Jessie openssl does not support proxies, so we can't specify a
        tor proxy directly.

        >>> from denova.os.user import whoami
        >>> if whoami() == 'root':
        ...     ok, __, __ = verify_certificate('google.com', 443)
        ...     ok
        ... else:
        ...     print(True)
        True

        >>> if whoami() == 'root':
        ...     ok, __, error_message = verify_certificate('DeNova.private.server', 443)
        ...     ok
        ...     error_message is None
        ... else:
        ...     print(False)
        ...     print(False)
        False
        False
    '''

    def extract_cert(response):

        cert = None

        i = response.find('-----BEGIN CERTIFICATE-----')
        if i > 0:
            temp_cert = response[i:]
            i = temp_cert.find('-----END CERTIFICATE-----')
            if i > 0:
                cert = temp_cert[:i+len('-----END CERTIFICATE-----')]

        return cert

    def verify_date(cert):

        ok = True
        error_message = None

        not_before, not_after = get_valid_dates(cert)
        try:
            after_date = datetime.strptime(not_after, '%b %d %H:%M:%S %Y %Z')
        except:   # 'bare except' because it catches more than "except Exception"
            try:
                after_date = datetime.strptime(not_after, '%b %d %H:%M:%S %Y %z')
            except:   # 'bare except' because it catches more than "except Exception"
                after_date = datetime.now() + timedelta(days=1)
        if after_date < datetime.now():
            ok = False
            error_message = f'{EXPIRED_CERT_ERR_MSG} {not_after}'

        return ok, error_message

    ok = True
    error_message = ''
    cert = None

    server = f'{hostname}:{port}'
    log(f'verify cert for {server}')

    if ca_certs_dir is None:
        ca_certs_dir = get_ca_certs_dir()
    log(f'ca certs dir: {ca_certs_dir}')

    try:
        # s_client waits for stdin after connecting, so we provide a short stdin
        # _in=' ' instead of _in='' because apparently sh does something like
        # checks '_in' instead of '_in is None'
        command_args = ['openssl', 's_client', '-connect', server, '-CApath', ca_certs_dir]
        result = run(*command_args)
        ##result = sh.openssl('s_client', '-connect', server, '-CApath', ca_certs_dir, _in=' ')

    except CalledProcessError as cpe:
        ##except sh.ErrorReturnCode as erc:
        ok = False
        try:
            stderr = cpe.stderr.strip()
            log(f'verify failed stderr:\n{stderr}')
            # parse 'connect: No route to host\nconnect:errno=22'
            # to 'connect: No route to host'
            error_message = stderr.split('\n')[0].strip()
        except:   # 'bare except' because it catches more than "except Exception"
            error_message = cpe # 'Unable to verify SSL certificate'
        log(cpe)

    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    else:
        result_stderr = result.stderr
        log(f'verify result stderr:\n{result_stderr}')
        lines = result_stderr.split('\n')
        return_code = None
        for line in lines:
            if line.startswith('verify return:'):
                return_code = line[len('verify return:'):]
            elif line.startswith('verify error:'):
                m = re.match('verify error:num=(\d+):(.*)', line)
                if m:
                    return_code = m.group(1)
                    error_message += m.group(2)
                else:
                    return_code = -1
                    error_message = line

        # get the certificate so we can do additional verification
        cert = extract_cert(result.stdout)

        # it seems like we're never able to verify the local issuer so ignore the error
        if return_code == '0' and error_message == 'unable to get local issuer certificate':
            error_message = None

        if error_message is not None and len(error_message) > 0:
            ok = False
            log(f'error verifying {hostname} certificate: {error_message}')
        else:
            error_message = None

        if return_code == '0' and error_message is None:
            ok, error_message = verify_date(cert)

    log(f'{server} cert ok: {ok}')

    return ok, cert, error_message

def get_issuer(cert):
    '''
        Get the issuer of an SSL certificate.
    '''
    issuer = None
    try:
        ##result = sh.openssl('x509', '-noout', '-issuer', _in=cert)
        result = run('openssl', 'x509', '-noout', '-issuer', cert)
        result_stdout = result.stdout
        m = re.match('issuer=(.*)', result_stdout)
        if m:
            issuer = m.group(1).strip()
        else:
            log(f'issuer result stdout: {result_stdout}')
            log(f'issuer result stderr: {result.stderr}')
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    return issuer

def get_issued_to(cert):
    '''
        Get to whom an SSL certificate was issued.
    '''
    issued_to = None
    try:
        result = run('openssl', 'x509', '-noout', '-subject', cert)
        result_stdout = result.stdout
        m = re.match('subject=(.*)', result_stdout)
        if m:
            issued_to = m.group(1).strip()
        else:
            log(f'issued_to result stdout: {result_stdout}')
            log(f'issued_to result stderr: {result.stderr}')
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    return issued_to

def get_valid_dates(cert):
    '''
        Get the dates an SSL certificate is valid.
    '''
    not_before = not_after = None
    try:
        result = run('openssl', 'x509', '-noout', '-dates', cert)
        result_stdout = result.stdout
        m = re.match('notBefore=(.*?)\nnotAfter=(.*)', result_stdout)
        if m:
            not_before = m.group(1).strip()
            not_after = m.group(2).strip()
        else:
            log(f'dates result stdout: {result_stdout}')
            log(f'dates result stderr: {result.stderr}')
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    return not_before, not_after

def get_hash(cert):
    '''
        Get the hash of an SSL certificate.
    '''
    cert_hash = None
    try:
        result = run('openssl', 'x509', '-noout', '-hash', cert)
        result_stdout = result.stdout
        cert_hash = result_stdout.strip()
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    return cert_hash

def get_fingerprint(cert):
    '''
        Get the MD5 fingerprint of an SSL certificate.
    '''
    fingerprint = None
    try:
        result = run('openssl', 'x509', '-noout', '-fingerprint', cert)
        result_stdout = result.stdout
        m = re.match('SHA1 Fingerprint=(.*)', result_stdout)
        if m:
            fingerprint = m.group(1).strip()
        else:
            log(f'fingerprint result stdout: {result_stdout}')
            log(f'fingerprint result stderr: {result.stderr}')
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())

    return fingerprint

def generate_certificate(
  domain, dirname, private_key_name=PRIVATE_KEY, public_cert_name=PUBLIC_CERT, name=None, days=365):
    '''
        Generate a self-signed SSL certficate.

        Writes the public cert to the file dirname/public_cert_name.
        Creates a dir dirname/private. Writes the private key to
        dirname/private/private_key_name.

        >>> generate_certificate('test.domain.com', '/tmp')
        >>> os.path.exists(os.path.join('/tmp', PUBLIC_CERT))
        True
        >>> os.path.exists(os.path.join('/tmp', 'private', PRIVATE_KEY))
        True
    '''

    if name is None:
        name = domain

    log(f'generate certificate for {name}')

    if not os.path.exists(dirname):
        os.makedirs(dirname)
        log(f'created {dirname}')

    private_dirname = os.path.join(dirname, 'private')
    if not os.path.exists(private_dirname):
        os.mkdir(private_dirname)
        log(f'created {private_dirname}')
    try:
        run('chown', 'root:ssl-cert', private_dirname)
    except:   # 'bare except' because it catches more than "except Exception"
        try:
            run('chown', ':ssl-cert', private_dirname)
        except:   # 'bare except' because it catches more than "except Exception"
            try:
                run('chown', 'root:root', private_dirname)
            except:   # 'bare except' because it catches more than "except Exception"
                pass
    run('chmod', 'go-rwx', private_dirname)

    delete_old_cert(domain, dirname, private_key_name, public_cert_name)
    gen_private_key(domain, dirname, private_key_name)
    gen_csr(domain, dirname, name, private_key_name)
    gen_cert(domain, dirname, private_key_name, public_cert_name, days)
    log(f'created certificate for {domain}')

def gen_private_key(domain, dirname, private_key_name):
    '''
        Generate an openssl private key for the domain.

        >>> from shutil import rmtree
        >>> if os.path.exists('/tmp/private'):
        ...    rmtree('/tmp/private')
        >>> gen_private_key('test.domain.com', '/tmp', PRIVATE_KEY)
        >>> os.path.exists(os.path.join('/tmp', 'private', PRIVATE_KEY))
        True
    '''

    log('generate private key')

    private_dir = os.path.join(dirname, 'private')
    if not os.path.exists(private_dir):
        os.makedirs(private_dir)

    private_key = os.path.join(private_dir, private_key_name)
    temp_private_key = f'{private_key}.tmp'

    responses = [
        (f'Enter pass phrase for {private_key}:', 'secret'),
        (f'Verifying - Enter pass phrase for {private_key}:', 'secret'),
        ]

    args = ['genrsa', '-aes256', '-out', private_key, '4096']
    #args = ['genpkey', '-out', private_key, '-outform', 'PEM', '-aes256', '-algorithm', 'rsa', '-pkeyopt', 'rsa_keygen_bits:4096']
    log('responding to generating key')
    responder = Responder(responses, 'openssl', *args).run()
    log('finished responding to generating key')
    assert os.path.exists(private_key), f'could not generate {private_key}'

    log('copying private key to temp key')
    run('cp', private_key, temp_private_key)
    responses = [(f'Enter pass phrase for {temp_private_key}:', 'secret')]
    args = ['rsa', '-in', temp_private_key, '-out', private_key]
    log('answering openssl questions')
    responder = Responder(responses, 'openssl', *args).run()
    log(f'responder: {responder}')
    os.remove(temp_private_key)

def gen_csr(domain, dirname, name, private_key_name):
    ''' Generate an openssl CSR for the domain. '''

    log('generate csr')

    private_key = os.path.join(dirname, 'private', private_key_name)
    csr = os.path.join(dirname, f'{domain}.csr')

    responses = [
        ('Country Name \(2 letter code\) \[AU\]:', '.'),
        ('State or Province Name \(full name\) \[Some-State\]:', '.'),
        ('Locality Name \(eg, city\) \[\]:', '.'),
        ('Organization Name \(eg, company\) \[Internet Widgits Pty Ltd\]:', name),
        ('Organizational Unit Name \(eg, section\) \[\]:', '.'),
        ('Common Name \(e\.g\. server FQDN or YOUR name\) \[\]:', domain),
        ('Email Address \[\]:', '.'),
        ('A challenge password \[\]:', ''),
        ('An optional company name \[\]:', ''),
        ]

    args = ['req', '-new', '-key', private_key, '-out', csr]
    Responder(responses, 'openssl', *args).run()
    assert os.path.exists(csr), f'could not generate {csr}'

def gen_cert(domain, dirname, private_key_name, public_cert_name, days):
    ''' Generate the public certificate. '''

    log('generating certificate')

    private_key = os.path.join(dirname, 'private', private_key_name)
    public_cert = os.path.join(dirname, public_cert_name)
    csr = os.path.join(dirname, f'{domain}.csr')

    run('openssl', 'x509','-req', '-days', str(days), '-in', csr, '-signkey', private_key, '-out', public_cert)
    assert os.path.exists(public_cert), f'could not generate {public_cert}'
    os.remove(csr)

    # only the owner should be able to read the private key
    run('chmod', 'u+r', private_key)
    run('chmod', 'u-wx', private_key)
    run('chmod', 'go-rwx', private_key)

    # everyone can read the public certificate
    run('chmod', 'ugo+r', public_cert)
    run('chmod', 'ugo-wx', public_cert)

def delete_old_cert(domain, dirname, private_key_name, public_cert_name):
    log(f'deleting old certficate files for {domain}')

    private_key = os.path.join(dirname, 'private', private_key_name)
    if os.path.exists(private_key):
        os.remove(private_key)
    elif os.path.exists(os.path.join(dirname, private_key_name)):
        os.remove(os.path.join(dirname, private_key_name))

    public_cert = os.path.join(dirname, public_cert_name)
    if os.path.exists(public_cert):
        os.remove(public_cert)

    csr = os.path.join(dirname, f'{domain}.csr')
    if os.path.exists(csr):
        os.remove(csr)

def move_private_key(dirname, private_key_name):
    ''' Move the private key to the dirname.

        By default openssl puts private keys in a 'private' subdir. Some
        apps need them in a different dir, such as the dir where the pub
        keys are.
    '''

    private_dir = os.path.join(dirname, 'private')
    private_key_path = os.path.join(private_dir, private_key_name)
    new_private_key_path = os.path.join(dirname, private_key_name)

    with locked():
        run('mv', private_key_path, new_private_key_path)
        log(f'moved {private_key_path} to {new_private_key_path}')
        # remove openssl's 'private' dir when empty
        if not os.listdir(private_dir):
            run('rmdir', private_dir)
        log(f'removed {os.path.join(private_dir, private_key_name)}')

def get_ca_certs_dir():
    '''
        Get the directory where openssl keeps known certs.

        >>> certs = get_ca_certs_dir()
        >>> certs == '/etc/ssl/certs'
        True
    '''

    # /etc/ssl/certs is debian standard ca certs dir
    # alternatives include /usr/lib/ssl/certs, etc.
    ca_certs_dir = '/etc/ssl/certs'
    """
    try:
        result = run('openssl', 'version', '-d')
        result_stdout = result.stdout
        m = re.match('OPENSSLDIR: "(.*?)"', result_stdout)
        if m:
            ca_certs_dir = '{}/certs'.format(m.group(1))
    except:   # 'bare except' because it catches more than "except Exception"
        log(format_exc())
    """

    return ca_certs_dir

def openssl_env():
    '''
        >>> env = openssl_env()
        >>> 'OPENSSL_CONF' in env
        True
    '''
    env = minimal_env()
    log(f'env: {env}')
    env['OPENSSL_CONF'] = OPENSSL_CONF_FILENAME
    log(f'env: {env}')

    return env

def sh_out(output):
    log.debug(output.rstrip())

def sh_err(output):
    log.warning(f'STDERR {output.rstrip()}')


if __name__ == "__main__":
    import doctest
    doctest.testmod()




'''
    Views

    Copyright 2019-2020 DeNova
    Last modified: 2020-10-24

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import os.path

try:
    import django
    django.setup()

    from django.http import Http404, HttpResponse
    from django.conf import settings
    from django.contrib.auth.views import PasswordChangeView, PasswordChangeDoneView
    from django.shortcuts import render
    from django.template import RequestContext

except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.python.log import get_log

log = get_log()


class ChangePassword(PasswordChangeView):
    ''' Change the password for regular user when logged in. '''
    # keep the following pass because when we strip comments, we need a line of code
    pass

class ChangePasswordDone(PasswordChangeDoneView):
    ''' Change the password done for regular user when logged in. '''
    # keep the following pass because when we strip comments, we need a line of code
    pass

def catch_all(request, url, context={}):
    ''' Handle urls not previously defined.

        Try to find a matching template file.

        Warning: Using this function in urls.py can mask earlier errors. (Why?)
                 If a page isn't working, try commenting out the line in urls.py.

        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('https://127.0.0.1/')
        >>> catch_all(request, 'https://127.0.0.1/unknown')
        Traceback (most recent call last):
             ...
        django.http.response.Http404: https://127.0.0.1/unknown
    '''

    def search_template_dirs(url):
        template = None
        for template_dir in settings.TEMPLATE_DIRS:
            if not template:
                path = os.path.join(template_dir, url)

                template = search_for_template(path, template_dir, url)
                if not template:
                    template = search_again(path, url)

        return template

    url = url.rstrip('/')
    log(f'unable to find {url} so making one last effort')

    template = search_template_dirs(url)
    if template is None:
        if url.endswith('index.html'):
            url = url.replace('index.html', 'home.html')
            template = search_template_dirs(url)
        elif url.endswith('index.htm'):
            url = url.replace('index.htm', 'home.html')
            template = search_template_dirs(url)

    for template_dir in settings.TEMPLATE_DIRS:
        if not template:
            path = os.path.join(template_dir, url)

            template = search_for_template(path, template_dir, url)
            if not template:
                template = search_again(path, url)

    if template:
        response = render(request,
                          template,
                          context=RequestContext(request, context).flatten())
    else:
        log.debug(f'raising 404 for {url}')
        raise Http404(url)

    return response

def search_for_template(path, template_dir, url):
    template = None

    if os.path.exists(path):
        if os.path.isdir(path):
            for home in ['home', 'index']:
                for ext in ['.html', '.htm']:
                    if not template:
                        basename = home + ext
                        path = os.path.join(template_dir, url, basename)
                        if os.path.exists(path):
                            template = os.path.join(url, basename)
        else:
            template = url

    return template

def search_again(path, url):
    template = None

    for ext in ['.html', '.htm']:
        if not template:
            if not path.endswith(ext):
                full_path = path + ext
                if os.path.exists(full_path):
                    template = url + ext

    return template

def empty_view(request, url):
    ''' Return an empty response, i.e. no html, for junk urls.

        Example: bootstrap.min.css.map[Learn More]


        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('https://127.0.0.1')
        >>> empty_view(request, 'https://127.0.0.1/bootstrap.min.css.map[Learn More]')
        <HttpResponse status_code=200, "text/html; charset=utf-8">
    '''

    log.debug(f'return an empty response for url: {url} with {request} request')

    return HttpResponse('')



'''
    Virtualenv without shell scripts.

    Copyright 2011-2020 DeNova
    Last modified: 2020-10-20

    No more shell script wrappers with "cd VIRTUALENV_DIR ; bin/activate".
    Just pure python scripts using virtualenv.

    Example::

        import ve
        ve.activate()

        import ... # from virtualenv

        ... code to run in virtualenv ...

    Example::

        from ve import venv

        with venv():
            import ... # from virtualenv
            ... code to run in virtualenv ...

        with venv(other_vdir):
            ... code to run in a different virtualenv ...

        ... code to run outside of virtualenv ...

    This module should be installed system wide, not inside a virtualenv.

    If no virtualenv directory is supplied to activate() or venv(),
    this module will search the calling module's dir and its
    parent dirs for a virtualenv. See virtualenv_dir().

    For maintainers: This module should never have any imports
    except from the standard python library. This allows you to import
    this module, activate a virtualenv, and then import other modules
    from that virtualenv.

    Do not import anything non-standard in this module at global scope.
    If you really need a non-standard import, import in a local scope.
    Example::

        _log = None

        def log(message):
            global _log
            if _log == None:
                # delayed non-global import, still takes effect globally
                import denova.python.log
                _log = denova.python.log.get()
            _log(message)

    This module is not named virtualenv because that module is part of the virtualenv
    program itself.

    To do:
        Check whether we are already in an activated virtualenv.

    ::

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import os
import sys
import traceback
from contextlib import contextmanager
from glob import glob

DEBUGGING = True
LOGGING = True

# site_packages_subdir_glob is relative to the virtualenv dir
site_packages_subdir_glob = 'lib/python*/site-packages'

if LOGGING:
    from denova.python.log import get_log
    log = get_log()

def debug(msg):
    if DEBUGGING:
        if LOGGING:
            log.debug(msg)
        else:
            print(msg)

@contextmanager
def venv(dirname=None, django_app=None, restore=True):
    ''' Context manager to activate a virtualenv.

        Example::

            from ve import venv

            with venv():

                # imports delayed until in virtualenv
                import ...

                ... code to run in virtualenv ...

        'dirname' is the directory where venv will start the search for
        a virtualenv. The default dirname is the current dir. The search
        includes dirname and its parent dirs.

        'django_app' is the django module with settings.py. Do not confuse
        this with a django subapp, also confusingly called a django app.

        To activate a virtualenv once for a module, use activate().

        This context manager will:
           * Set the VIRTUAL_ENV environment variable to the virtualenv dir
           * Set the current dir to the virtualenv dir
           * Prepend virtualenv/bin to the os environment PATH
           * Prepend sites-packages to sys.path
           * Optionally set the DJANGO_SETTINGS_MODULE environment variable

        If dir is not included or is None, ve searches for a virtualenv.
        See virtualenv_dir().

       On exiting the context, any changes to these system environment
       variables or python's sys.path are lost.

        Virtualenv's 'bin/activate' doesn't work well with fabric. See
        http://stackoverflow.com/questions/1691076/activate-virtualenv-via-os-system
    '''

    # venv is called from denova/wsgi_django.py every time
    # gunicorn restarts (starts?) a worker # DEBUG
    # from denova.python.utils import stacktrace # DEBUG
    # debug('called venv() from:\n{}'.format(stacktrace())) # DEBUG

    old_virtualenv = os.environ.get('VIRTUAL_ENV')
    old_settings_module = os.environ.get('DJANGO_SETTINGS_MODULE')
    old_path = os.environ.get('PATH')
    old_python_path = list(sys.path)
    try:
        old_cwd = os.getcwd()
    except:   # 'bare except' because it catches more than "except Exception"
        # import late so environment is configured properly
        from denova.os.user import getdir

        old_cwd = getdir()

    if dirname:
        debug(f'starting dirname: {dirname}')
    else:
        debug(f'set dirname to cwd: {old_cwd}')
        dirname = old_cwd

    venv_dir = virtualenv_dir(dirname)
    debug(f'set up virtual_env: {dirname}')

    os.environ['VIRTUAL_ENV'] = venv_dir

    bin_dir = os.path.join(venv_dir, 'bin')
    path_dirs = os.environ['PATH'].split(':')
    if bin_dir not in path_dirs:
        new_path = ':'.join([bin_dir] + path_dirs)
        os.environ['PATH'] = new_path

    if django_app:
        os.environ['DJANGO_SETTINGS_MODULE'] = f'{django_app}.settings'
        debug(f'django settings: {django_app}')

    os.chdir(venv_dir)

    sys.path = venv_sys_path(venv_dir)

    try:
        yield

    finally:

        # activate() isn't a context manager, so it sets restore=False
        if restore:

            debug('restore')
            if old_virtualenv:
                os.environ['VIRTUAL_ENV'] = old_virtualenv
            else:
                del os.environ['VIRTUAL_ENV']

            os.environ['PATH'] = old_path

            if old_settings_module:
                os.environ['DJANGO_SETTINGS_MODULE'] = old_settings_module
            else:
                if 'DJANGO_SETTINGS_MODULE' in os.environ:
                    del os.environ['DJANGO_SETTINGS_MODULE']

            try_to_cd_back(old_cwd)

            sys.path[:] = old_python_path

    debug('finished venv()')

def activate(dirname=None, django_app=None):
    ''' Activate a virtualenv.

        Example::

            # before any imports from a virtualenv
            import ve
            ve.activate(dirname)   # or ve.activate()

            # now we can import from the virtualenv
            import ...

        If dirname is not included or is None, ve searches for a virtualenv.
        See virtualenv_dir().

        If you want to enter and then exit the virtualenv, use the context
        manager venv().
    '''

    venv(dirname, django_app=django_app, restore=False).__enter__()

def in_virtualenv(dirname=None):
    ''' Return True if in virtualenv, else return False.

        If dirname is specified, return if in specified venv. '''

    """ This is the method used internally in virtualenv.
        See https://stackoverflow.com/questions/1871549/determine-if-python-is-running-inside-virtualenv """

    in_venv = False

    if 'VIRTUAL_ENV' in os.environ:

        if not dirname:
            dirname = os.environ['VIRTUAL_ENV']

        if dirname == os.environ['VIRTUAL_ENV']:
            bin_dir = os.path.join(dirname, 'bin')
            path_dirs = os.environ['PATH'].split(':')

            if bin_dir in path_dirs:

                if os.path.isdir(bin_dir):
                    in_venv = True

    return in_venv

def virtualenv_dir(dirname=None):
    ''' Return full path to virtualenv dir. Raises exception if the
        specified dirname is not a virtualenv and no virtualenv is found.

        virtualenv_dir() searches for a virtualenv in:
            * dirname
            * any parent dir of dirname
            * any immediate subdir of the above

        If dirname is None (the default), virtualenv_dir() sets it to the
        first of:
            * The VIRTUAL_ENV environment variable
            * a virtualenv in a calling module's dir, or parent dir

        This lets you run a program which automatically finds and then
        activates its own virtualenv. You don't need a wrapper script to
        first activate the virtualenv and then run your python code.

        A virtualenv dir is a dir containing "bin/python", "bin/activate", and
        "bin/pip".

        Since there is probably no virtualenv associated with a dir in the
        system path, links in the PATH env dirs are followed. So for a
        program in a PATH dir which is a link, its dirname is the dir of
        the target of the link.

        If you have a default virtualenv dir that is not in your module's
        directory tree, you can still have ve automatically find it. Create
        a link to the virtualenv dir in the calling module's dir or one of
        its parent dirs.

        Example::

            # if your default virtualenv dir is mydefault/virtualenv
            # and your python code is in /usr/local/bin
            ln --symbolic mydefault/virtualenv /usr/local/bin

        '''

    if dirname:
        vdir = check_dir_for_virtualenv(dirname)

    elif 'VIRTUAL_ENV' in os.environ:
        base_dirname = os.environ['VIRTUAL_ENV']
        vdir = os.path.abspath(base_dirname)

    else:

        vdir = None

        # find the calling program module
        # we want the last caller which is not this module
        stack = traceback.extract_stack()
        debug('stack: {}'.format('\n'+'\n'.join(repr(element) for element in stack)))
        debug(f'__file__: {__file__}')

        # stack filenames end in .py; __file__ filenames end in .pyc or .pyo
        basename, _, extension = __file__.rpartition('.')
        if extension == 'pyc' or extension == 'pyo':
            this_filename = basename + '.py'
        else:
            this_filename = __file__
        debug(f'basename: {basename}')
        debug(f'extension: {extension}')
        debug(f'this_filename: {this_filename}')

        caller_filenames = []
        found_callers = False
        for filename, line_number, function_name, text in stack:
            if not found_callers:
                if filename == this_filename:
                    found_callers = True
                else:
                    #debug('caller filename: {}'.format(filename))
                    caller_filenames.append(filename)

        # do we want to start the search at the earliest or most recent caller?
        # i.e. do we want to leave the list as is or reverse it?
        # for now most recenty, and so we reverse the list
        debug('reverse caller filenames')
        caller_filenames = reversed(caller_filenames)
        for caller_filename in caller_filenames:
            #debug('(reversed) caller filename: {}'.format(caller_filename))
            if not vdir:
                """ what's this about?
                # in PATH, follow links
                # we probably don't want to always follow links
                path = os.environ['PATH'].split(':')
                dirname = os.path.dirname(caller_filename)
                debug('os.path.islink({}): {}'.format(caller_filename, os.path.islink(caller_filename)))
                while os.path.islink(caller_filename) and dirname in path:
                    caller_filename = os.readlink(caller_filename)
                """

                dirname = os.path.dirname(caller_filename)
                debug(f'dirname: {dirname}')
                vdir = check_dir_for_virtualenv(dirname)

    if not vdir:
        raise Exception(f'No virtualenv found for {dirname}')

    return vdir

def check_dir_for_virtualenv(dirname):
    ''' Check dir and its parent dirs for a virtualenv. '''

    dirname = os.path.abspath(dirname)

    vdir = None
    done = False
    while not done:

        #debug('check_dir_for_virtualenv() dirname: {}'.format(dirname))
        if not dirname:
            debug('done because no dirname')
            done = True

        elif dirname == '/':
            debug('done dirname is /')
            done = True

        elif is_virtualenv(dirname):
            debug(f'done because virtualenv found: {dirname}')
            vdir = dirname
            done = True

        else:
            #debug('check_dir_for_virtualenv() checking subdirs of : {}'.format(dirname))
            for d in os.listdir(dirname):
                if not done:
                    path = os.path.join(dirname, d)
                    if os.path.isdir(path):
                        if is_virtualenv(path):
                            vdir = path
                            done = True
            if done: debug(f'vdir: {vdir}')

        if not done:
            # try the parent dir
            dirname = os.path.dirname(dirname)

    if vdir:
        debug(f'vdir found: {vdir}')

    return vdir


def is_virtualenv(vdir):
    ''' Return whether specified dir is a virtualenv. '''

    return (
        vdir and
        os.path.exists(os.path.join(vdir, 'bin', 'python')) and
        os.path.exists(os.path.join(vdir, 'bin', 'activate')) and
        os.path.exists(os.path.join(vdir, 'bin', 'pip'))
        )

def venv_sys_path(dirname=None):
    ''' Return sys.path for venv dir '''

    venv_dir = virtualenv_dir(dirname)

    # import late so environment is configured
    from denova.os.command import run

    old_dir = os.getcwd()
    os.chdir(venv_dir)

    # get the sys.path from the virtual env's python
    # ?? should we just use the venv's default bin/python?
    major, _, _, _, _ = sys.version_info
    python_path = os.path.join(venv_dir, f'bin/python{major}')
    result = run(python_path, '-c', 'import sys ; print(sys.path)')
    sys_path_string = result.stdout
    result = eval(sys_path_string)

    try_to_cd_back(old_dir)

    return result

def site_packages_dir(dirname=None):
    ''' Return site-packages dir for venv dir '''

    venv_dir = virtualenv_dir(dirname)

    old_dir = os.getcwd()
    os.chdir(venv_dir)

    local_site_packages_dir = glob(site_packages_subdir_glob)[0]
    result = os.path.abspath(os.path.join(venv_dir, local_site_packages_dir))

    try_to_cd_back(old_dir)

    return result

def package_dir(package, dirname=None):
    ''' Return package dir in venv dir '''

    return os.path.join(site_packages_dir(dirname), package)

def try_to_cd_back(old_cwd):
    # because we may have su'd to another user since we originally
    # cd'd to the old dir, we may not have permission to cd back
    try:
        os.chdir(old_cwd)
    except OSError:
        # just log it
        #log('could not chdir(%s); probably not an error' % old_cwd)
        pass


if __name__ == "__main__":

    import doctest
    doctest.testmod()



'''
    Lookup key in dict

    Usage:

        {{ dictionary|lookup:key }}

    From:
        https://code.djangoproject.com/ticket/12486
        http://push.cx/2007/django-template-tag-for-dictionary-access

    Last modified: 2020-10-01
'''

try:
    from django import template
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')


register = template.Library()

@register.filter
def lookup(d, key):
    ''' Fix for django's lack of variable dict access in templates.

        Usage: {{ dictionary|lookup:key }}

        In a template, "{{ mydict.value }}" means "mydict['value']"
        instead of "mydict[value]".

        Use "{% for key, value in mydict.iteritems %}" when you can. But
        sometimes you can't, such as when you want dict access in a specific
        order.
    '''

    return d[key]



'''
    Template middleware.

    Copyright 2011-2020 DeNova
    Last modified: 2020-10-01

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

try:
    from django.conf import settings
    from django.utils.deprecation import MiddlewareMixin
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.python.log import get_log

log = get_log()

def set_context(response, key, value):
    # is this necessary, or can we just e.g. "response.context_data['request'] = request"?
    try:
        # TemplateResponse
        context = response.context
    except AttributeError:
        # SimpleTemplateResponse
        context = response.context_data
    if context is None:
        context = {}

    if key not in context:
        context[key] = value

class RequestMiddleware(MiddlewareMixin):
    ''' Add the request to the template context. '''

    def process_template_response(self, request, response):

        set_context(response, 'request', request)
        return response

class SettingsMiddleware(MiddlewareMixin):
    ''' Add the settings to the template context. '''

    def process_template_response(self, request, response):

        set_context(response, 'settings', settings)
        return response



'''
    Multiprocess-safe locks.

    Requires the safelog package available on PyPI.

    Copyright 2011-2020 DeNova
    Last modified: 2020-10-24

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import os
import socket
import sys
import threading
from contextlib import contextmanager
from datetime import timedelta
from functools import wraps
from time import sleep
from traceback import format_exc

from denova.python.log import get_log
from denova.python.times import now
from denova.python.utils import caller_id, object_name

# constants shared with denova.python.log and denova.python.logwriter are
# in denova.python.log so they can be imported easily by tools
LOCK_SERVER_HOST = 'localhost'
LOCK_SERVER_PORT = 8674
BIND_ADDR = (LOCK_SERVER_HOST, LOCK_SERVER_PORT)
MAX_PACKET_SIZE = 1024

ACTION_KEY = 'action'
LOCKNAME_KEY = 'lockname'
NONCE_KEY = 'nonce'
PID_KEY = 'pid'
OK_KEY = 'ok'
MESSAGE_KEY = 'message'
LOCK_ACTION = 'lock'
UNLOCK_ACTION = 'unlock'

REJECTED_LOCK_MESSAGE = 'Lockserver rejected "{}" lock request: {}'
REJECTED_UNLOCK_MESSAGE = 'Lockserver rejected "{}" unlock request: {}'
WHY_UNKNOWN = 'lockserver did not say why'

DEFAULT_TIMEOUT = timedelta.max

# global variables
log = get_log()
# WARNING: BUG. python globals are not multiprocess-safe.
synchronized_locks = {}

class LockTimeout(Exception):
    pass

class LockFailed(Exception):
    pass

@contextmanager
def locked(lockname=None, timeout=None):
    ''' Get a simple reusable lock as a context manager.

        'name' same as lock(). Default is a name created from the
        calling module and line number.

        'timeout' is the maximum time locked() waits for a lock,
        as a  timedelta. Default is one minute.
        If a lock waits longer than 'timeout',
        locked() logs the failure and raises LockTimeout.
        If your locked code block can take longer, you must set
        'timeout' to the longest expected time.

        With locked() you don't have to initialize each lock in an
        an outer scope, as you do with raw multiprocessing.Lock().

        The lock returned by locked() is also a context manager. You
        don't have to explicitly call acquire() or release().

        >>> with locked():
        ...     print('this is a locked code block')
        this is a locked code block

        >>> with locked(timeout=1):
        ...     print('this is a locked code block with a timeout')
        ...     sleep(2)
        this is a locked code block with a timeout
        >>> print('after locked code block with a timeout')
        after locked code block with a timeout

        The python standard multiprocessing. Lock won't let you call
        multiprocessing.Lock.release() if the lock is already unlocked. The
        context manager returned by locked() enforces that restriction
        painlessly by calling release() automatically for you.

        If for some reason you use 'with locked()' with no name twice on the
        same line, it will return the same lock twice. You're extremely
        unlikely to do that accidentally.
    '''

    try:
        # log('enter locked() {}'.format(lockname)) # DEBUG

        if not lockname:
            lockname = caller_id(ignore=[__file__, r'.*/contextlib.py'])
        # log('lockname: {}'.format(lockname)) # DEBUG
        warning_msg = f'lock timed out called from {lockname}'
        log(warning_msg)

        # log('call lock({})'.format(lockname)) # DEBUG
        is_locked, nonce, pid = lock(lockname, timeout)

    except Exception as exc:
        #from denova.python.utils import stacktrace

        log(f'Unexpected exception: {str(exc)}')
        # don't stop if running test_until_exception
        # error_message = stacktrace().replace('Traceback', 'Stacktrace') # but without 'Traceback' # DEBUG
        # log(error_message)
        # DEBUG log(format_exc())
        raise

    else:
        try:
            yield
        finally:
            unlock(lockname, nonce, pid, timeout)

    # log('exit locked() {}'.format(lockname)) # DEBUG

def lock(lockname, timeout=None):
    '''
        Lock a process or thread to prevent concurrency issues.

        'lockname' is the name of the lock.

        Every process or thread that calls "lock()" from the
        same source file and line number contends for the same
        lock. If you want many instances of a class to run at
        the same time, each instance's lockname for a particular
        call to lock() must use a different lockname.
        Example::

            lockname = 'MyClass {}'.format(self.instance_id())
            lock(lockname)

        You may still choose to include the source path and line number
        from denova.python.process.caller() in your lockname.

        If for some reason you use 'with locked()' with no name twice on the
        same line, the second 'with locked()' will fail. They both have the
        same default lockname with the same caller and line number. You're
        extremely unlikely to do that accidentally.

        >>> pid = os.getpid()
        >>> log('pid: {}'.format(pid))

        >>> log('test simple lock()/unlock()')
        >>> from denova.os.process import is_pid_active
        >>> lockname = 'lock1'
        >>> is_locked, nonce, pid = lock(lockname)
        >>> is_locked
        True
        >>> isinstance(nonce, str)
        True
        >>> is_pid_active(pid)
        True
        >>> unlock(lockname, nonce, pid)
        True

        >>> log('test relock')
        >>> lockname = 'lock1'
        >>> is_locked, nonce, __ = lock(lockname)
        >>> is_locked
        True

        >>> log('while locked, try to lock again should fail')
        >>> try:
        ...     lock(lockname, timeout=timedelta(milliseconds=3))
        ... except LockTimeout as lt:
        ...     print(str(lt))
        lock timed out: lock1

        >>> log('now unlock it')
        >>> unlock(lockname, nonce, pid)
        True

        >>> log('try 2 locks')
        >>> lockname1 = 'lock1'
        >>> is_locked1, nonce1, pid1 = lock(lockname1)
        >>> is_locked1
        True
        >>> lockname2 = 'lock2'
        >>> is_locked2, nonce2, pid2 = lock(lockname2)
        >>> is_locked2
        True
        >>> nonce1 != nonce2
        True
        >>> pid1 == pid2
        True
        >>> unlock(lockname1, nonce1, pid1)
        True
        >>> unlock(lockname2, nonce2, pid2)
        True
    '''

    nonce = None
    pid = os.getpid()

    deadline = get_deadline(timeout)
    # log('lock deadline: {}'.format(deadline)) # DEBUG

    # we can probably factor this out into a general case
    loop_count = 0
    is_locked = False
    last_warning = None
    while not is_locked:
        try:
            # only report every 10 secs
            # if (loop_count % 100) == 0:
                # log('call lock({})'.format(lockname)) # DEBUG

            is_locked, nonce = try_to_lock(lockname, pid)

        except TimeoutError as te:
            log(str(te))

        except LockFailed as lf:
            # we need a better way to handle serious errors
            if 'Wrong nonce' in str(lf):
                raise
            else:
                message = lf.args[0]
                if message != last_warning:
                    last_warning = message
                    log(message)

        except:   # 'bare except' because it catches more than "except Exception"
            log(format_exc())
            raise

        if not is_locked:
            if deadline and now() > deadline:
                warning_msg = f'lock timed out: {lockname}'
                log.warning(warning_msg)
                raise LockTimeout(warning_msg)

            sleep(0.1)

        loop_count = loop_count + 1

    return is_locked, nonce, pid

def try_to_lock(lockname, pid):
    ''' Try once to lock. '''

    is_locked = False
    nonce = None

    # Create a socket (SOCK_STREAM means a TCP socket)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:

        connect_to_server(sock)

        # send request
        data = {ACTION_KEY: LOCK_ACTION,
                LOCKNAME_KEY: lockname,
                PID_KEY: pid}
        # log('about to send lock request: {}'.format(data))
        sock.sendall(repr(data).encode())
        # log('finished sending lock request')

        # get response
        data = sock.recv(MAX_PACKET_SIZE)
        # log('finished receiving lock data: {}'.format(data))
        response = eval(data.decode())

        is_locked = (response[OK_KEY] and
                     response[ACTION_KEY] == LOCK_ACTION and
                     response[LOCKNAME_KEY] == lockname)

        if is_locked:
            nonce = response[NONCE_KEY]
            # log('locked: {} with {} nonce'.format(lockname, nonce)) # DEBUG
        else:
            # if the server responded with 'No'
            if MESSAGE_KEY in response:
                message = REJECTED_LOCK_MESSAGE.format(lockname, response[MESSAGE_KEY])
            else:
                message = REJECTED_LOCK_MESSAGE.format(lockname, WHY_UNKNOWN)
            #log(message)
            raise LockFailed(message)

    return is_locked, nonce

def unlock(lockname, nonce, pid, timeout=None):
    '''
        >>> log('Unlock a process or thread that was previously locked.')
        >>> lockname = 'lock1'
        >>> __, nonce, pid = lock(lockname)
        >>> unlock(lockname, nonce, pid)
        True

        >>> log('A bad nonce should fail.')
        >>> lockname = 'lock1'
        >>> __, nonce, pid = lock(lockname)
        >>> try:
        ...    unlock(lockname, 'bad nonce', pid)
        ...    assert False, 'Unexpectedly passed bad nonce'
        ... except LockFailed:
        ...     pass
        >>> unlock(lockname, nonce, pid)
        True
    '''

    deadline = get_deadline(timeout)
    # log('unlock deadline: {}'.format(deadline)) # DEBUG

    # we must be persistent in case the lockserver is busy
    is_locked = True
    last_warning = None
    while is_locked:
        try:
            is_locked = try_to_unlock(lockname, nonce, pid)

        except TimeoutError as te:
            log(str(te))

        except LockFailed as lf:
            # we need a better way to handle serious errors
            if 'Wrong nonce' in str(lf):
                raise
            else:
                message = lf.args[0]
                if message != last_warning:
                    last_warning = message
                    log(message)

        except:   # 'bare except' because it catches more than "except Exception"
            log(format_exc())
            raise

        if is_locked:
            if deadline and now() > deadline:
                warning_msg = f'unlock timed out: {lockname}'
                log.warning(warning_msg)
                raise LockTimeout(warning_msg)

            sleep(0.1)

    # log('unlocked: {}'.format(lockname)) # DEBUG

    # only returned for testing purposes
    return not is_locked

def try_to_unlock(lockname, nonce, pid):
    ''' Try once to unlock. '''

    is_locked = True

    # Create a socket (SOCK_STREAM means a TCP socket)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:

        connect_to_server(sock)

        # send request
        data = {ACTION_KEY: UNLOCK_ACTION,
                LOCKNAME_KEY: lockname,
                NONCE_KEY: nonce,
                PID_KEY: pid}
        # log('about to send unlock request: {}'.format(data))
        sock.sendall(repr(data).encode())
        # log('finished sending unlock request')

        # get response
        data = sock.recv(MAX_PACKET_SIZE)
        # log('finished receiving unlock data: {}'.format(data))

        response = eval(data.decode())
        if response[OK_KEY] and response[ACTION_KEY] == UNLOCK_ACTION and response[NONCE_KEY] == nonce:

            is_locked = False

        else:
            # if the server responded with 'No'
            if MESSAGE_KEY in response:
                message = REJECTED_UNLOCK_MESSAGE.format(lockname, response[MESSAGE_KEY])
            else:
                message = REJECTED_UNLOCK_MESSAGE.format(lockname, WHY_UNKNOWN)
            #log(message)
            raise LockFailed(message)

    return is_locked

def synchronized(function):
    ''' Decorator to lock a function so each call completes before
        another call starts.

        If you use both the staticmethod and synchronized decorators,
        @staticmethod must come before @synchronized.
    '''

    @wraps(function)
    def synchronizer(*args, **kwargs):
        ''' Lock function access so only one call at a time is active.'''

        # get a shared lock for the function
        with locked():
            lock_name = object_name(function)
            if lock_name in synchronized_locks:
                synch_lock = synchronized_locks[lock_name]
            else:
                synch_lock = threading.Lock()
                synchronized_locks[lock_name] = synch_lock

        with locked():
            result = function(*args, **kwargs)

        return result

    return synchronizer

def get_deadline(timeout=None):
    '''
        Return a datetime deadline from timeout.

        'timeout' can be seconds or a timedelta. Default is timedelta.max.

        >>> from datetime import datetime
        >>> deadline = get_deadline()
        >>> deadline is None
        True

        >>> deadline = get_deadline(timedelta(seconds=1))
        >>> type(deadline) is datetime
        True

        >>> deadline = get_deadline(1)
        >>> type(deadline) is datetime
        True

        >>> deadline = get_deadline(1.1)
        >>> type(deadline) is datetime
        True

        >>> deadline('bad timeout value')
        Traceback (most recent call last):
        ...
        TypeError: 'datetime.datetime' object is not callable
    '''

    if timeout is None:
        deadline = None
    elif isinstance(timeout, timedelta):
        deadline = now() + timeout
    elif isinstance(timeout, (float, int)):
        deadline = now() + timedelta(seconds=timeout)
    else:
        raise ValueError(f'timeout must be one of (seconds, timedelta, None), not {type(timeout)}')

    # log('deadline: {}'.format(deadline))
    return deadline

def connect_to_server(sock):
    ''' Connect to lockserver. '''

    try:
        sock.connect(BIND_ADDR)

    except ConnectionRefusedError:
        msg = f'Requires DeNova lockserver. No lock server at {LOCK_SERVER_HOST}:{LOCK_SERVER_PORT}'
        print(msg, file=sys.stderr)
        log(msg)
        raise LockFailed(msg)

if __name__ == "__main__":

    import doctest
    doctest.testmod()



'''
    Alternate log. Primarily for denova.python.log and modules it uses.

    Copyright 2015-2020 DeNova
    Last modified: 2020-10-21

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.

    Logs to the file _log.USER.log in the dir returned by tempfile.gettempdir().

    The modules denova.python.log uses cannot use denova.python.log itself. Use denova.python.log instead
    of this module if you can. This module is much less efficient and
    powerful than denova.python.log. To debug this module use print().

    Functions that are used by both log and _log are here.
'''

import os
import pwd
import time
from tempfile import gettempdir
from traceback import format_exc

def log(message, filename=None, mode=None):
    ''' Log message that denova.python.log can't. '''

    if filename is None:
        try:
            user = whoami()
        # we want to catch any type of exception and Exception won't always do that
        except:
            user = 'unknown'
        filename = os.path.join(gettempdir(), f'_log.{user}.log')
    if mode is None:
        mode = '0666'

    with open(filename, 'a') as logfile:
        current_timestamp = timestamp()
        try:
            # logwriter dies sometimes and stops regular logging
            # but logit itself logs to this alternate log
            # this print should goto the systemd journal
            #    journalctl --unit logit
            # error is e.g.:
            #    2020-04-17 19:21:33,555 too many values to unpack (expected 3)
            logfile.write(f'{current_timestamp} {format_exc()}\n') # DEBUG
            logfile.write(f'{current_timestamp} {message}\n')
        except UnicodeDecodeError:
            from denova.python.utils import is_string

            try:
                logfile.write(f'unable to write message because it is a type: {type(message)}')
                if not is_string(message):
                    decoded_message = message.decode(errors='replace')
                    logfile.write(f'{current_timestamp} {decoded_message}\n')

            except: # 'bare except' because it catches more than "except Exception"
                print(format_exc())


# redir log.debug() etc.
log.debug = log
log.warning = log
log.error = log

def whoami():
    ''' Get user '''

    # do not use denova.os.user.whoami() so we avoid circular imports
    return pwd.getpwuid(os.geteuid()).pw_name

def timestamp():
    ''' Timestamp as a string. Duplicated in this module to avoid recursive
        imports. '''

    current_time = time.time()
    milliseconds = int((current_time - int(current_time)) * 1000)
    formatted_current_time = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
    formatted_milliseconds = '{:03}'.format(milliseconds)
    return f'{formatted_current_time},{formatted_milliseconds}'



'''
    Convert an image file to a data uri.

    Copyright 2012-2020 DeNova
    Last modified: 2020-10-10

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

try:
    from django import template
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

import denova.django_addons.data_image

register = template.Library()

@register.filter
def data_img(filename, browser=None):
    ''' Encode an image file in base 64 as a data uri.
        The filename is relative to settings.STATIC_URL/settings.STATIC_ROOT.

        If the data uri is too large or anything goes wrong,
        returns the static path to the image file.

        Example:

            <img alt="embedded image" src="{{ 'images/myimage.png'|data_img:browser }}">

    '''

    return denova.django_addons.data_image.data_image(filename, browser=browser)



"""
    Python CSS Browser Selector v0.0.1
    Koes Bong (http://web.koesbong.com)
    http://web.koesbong.com/2011/01/28/python-css-browser-selector/
    License: http://creativecommons.org/licenses/by/2.5/
    Credits: This is a python port from Bastian Allgeier's PHP CSS Browser Selector:
    http://www.bastian-allgeier.de/css_browser_selector/,
    which is a port from Rafael Lima's original Javascript CSS Browser Selector:
    http://rafael.adm.br/css_browser_selector
"""

import re

def get_ua(ua):

    ua = ua.lower()
    b = []

    # browser
    get_standardized_browser(ua, b)

    #platform
    get_standardized_platform(ua, b)

    css_selector = {"css": " " . join(b)}

    return css_selector

def get_standardized_browser(ua, b):

    g = "gecko"
    w = "webkit"
    s = "safari"

    opera_webtv_matches = re.search(r"opera|webtv", ua)
    opera_matches = re.search(r"opera(\s|\/)(\d+)", ua)
    msie_matches = re.search(r"msie\s(\d)", ua)

    if opera_webtv_matches is None and msie_matches is not None:
        b.append("ie ie" + msie_matches.group(1))
    elif ua.find(r"firefox/2") != -1:
        b.append(g + " ff2")
    elif ua.find(r"firefox/4") != -1:
        b.append(g + " ff4")
    elif ua.find(r"firefox/3.6") != -1:
        b.append(g + " ff36")
    elif ua.find(r"firefox/3.5") != -1:
        b.append(g + " ff35")
    elif ua.find(r"firefox/3") != -1:
        b.append(g + " ff3")
    elif ua.find(r"firefox/5") != -1:
        b.append(g + " ff5")
    elif ua.find(r"gecko/") != -1:
        b.append(g)
    elif opera_matches is not None:
        b.append("opera opera" + opera_matches.group(2))
    elif ua.find(r"konquerer") != -1:
        b.append("konquerer")
    elif ua.find(r"chrome") != -1:
        b.append(w + " " + s + " chrome")
    elif ua.find(r"iron") != -1:
        b.append(w + " " + s + " iron")
    elif ua.find(r"applewebkit/") != -1:
        applewebkit_ver_matches = re.search(r"version\/(\d+)", ua)
        if applewebkit_ver_matches is not None:
            b.append(w + " " + s + " " + s + applewebkit_ver_matches.group(1))
        else:
            b.append(w + " " + s)
    elif ua.find(r"mozilla/") != -1:
        b.append(g)
    elif ua.find(r"unknown"):
        b.append("unknown")

def get_standardized_platform(ua, b):
    if ua.find("j2me") != -1:
        b.append("j2me")
    elif ua.find("iphone") != -1:
        b.append("iphone")
    elif ua.find("ipod") != -1:
        b.append("ipod")
    elif ua.find("ipad") != -1:
        b.append("ipad")
    elif ua.find("android") != -1:
        b.append("android")
    elif ua.find("blackberry") != -1:
        b.append("blackberry")
    elif ua.find("mobile") != -1:
        b.append("mobile")
    elif ua.find("mac") != -1 or ua.find('darwin') != -1:
        b.append("mac")
    elif ua.find("webtv") != -1:
        b.append("webtv")
    elif ua.find("win") != -1:
        b.append("win")
    elif ua.find("freebsd") != 1:
        b.append("freebsd")
    elif ua.find("x11") != -1 or ua.find("linux") != -1:
        b.append("linux")



#!/usr/bin/env python3
'''
    Browser utilities.

    Copyright 2012-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import re
from traceback import format_exc

from denova.python.log import get_log

log = get_log()

def user_agent_tags(ua):
    ''' Returns list of browser types indicated by the user agent.

        License: http://creativecommons.org/licenses/by/2.5/
        Credits:
            DeNova: http://DeNova.com
            This is a port from Koes Bong: http://web.koesbong.com/2011/01/28/python-css-browser-selector/,
            which is a port from Bastian Allgeier's PHP CSS Browser Selector: http://www.bastian-allgeier.de/css_browser_selector/,
            which is a port from Rafael Lima's original Javascript CSS Browser Selector: http://rafael.adm.br/css_browser_selector

    >>> user_agent_tags('Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/536.5 (KHTML, like Gecko) Chrome/19.0.1084.56 Safari/536.5')
    ['webkit safari chrome', 'linux']
    >>> user_agent_tags('Safari/7534.57.2 CFNetwork/520.4.3 Darwin/11.4.0 (x86_64) (MacBookPro8%2C2)')
    ['unknown', 'mac']
    >>> user_agent_tags('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; .NET CLR 1.1.4322; .NET CLR 2.0.50727; .NET CLR 3.0.04506.30)')
    ['ie ie7', 'win']
    >>> user_agent_tags('Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.0.19) Gecko/2012020109 Iceweasel/3.0.6 (Debian-3.0.6-3)')
    ['gecko', 'linux']
    >>> user_agent_tags('curl/7.21.0 (i486-pc-linux-gnu) libcurl/7.21.0 OpenSSL/0.9.8o zlib/1.2.3.4 libidn/1.15 libssh2/1.2.6')
    ['curl', 'linux']
    >>> user_agent_tags('Lynx/2.8.8dev.5 libwww-FM/2.14 SSL-MM/1.4.1 GNUTLS/2.8.6')
    ['lynx']
    >>> user_agent_tags('Wget/1.12 (linux-gnu)')
    ['wget', 'linux']

    '''

    browser = []

    try:
        ua = ua.lower()
        analyze_browser_type(ua, browser)
        analyze_platform(ua, browser)

    except: # 'bare except' because it catches more than "except Exception"
        browser.append('unknown')
        log(format_exc())

    return browser

def analyze_browser_type(ua, browser):
    '''
        Determine the browser type from the user agent.
    '''

    g = 'gecko'
    w = 'webkit'
    s = 'safari'

    opera_webtv_matches = re.search(r'opera|webtv', ua)
    opera_matches = re.search(r'opera(\s|\/)(\d+)', ua)
    msie_matches = re.search(r'msie\s(\d)', ua)

    if opera_webtv_matches is None and msie_matches is not None:
        browser.append('ie ie' + msie_matches.group(1))
    elif ua.find(r'firefox/2') != -1:
        browser.append(g + ' ff2')
    elif ua.find(r'firefox/4') != -1:
        browser.append(g + ' ff4')
    elif ua.find(r'firefox/3.6') != -1:
        browser.append(g + ' ff36')
    elif ua.find(r'firefox/3.5') != -1:
        browser.append(g + ' ff35')
    elif ua.find(r'firefox/3') != -1:
        browser.append(g + ' ff3')
    elif ua.find(r'firefox/5') != -1:
        browser.append(g + ' ff5')
    elif ua.find(r'gecko/') != -1:
        browser.append(g)
    elif opera_matches is not None:
        browser.append('opera opera' + opera_matches.group(2))
    elif ua.find(r'konquerer') != -1:
        browser.append('konquerer')
    elif ua.find(r'chrome') != -1:
        browser.append(w + ' ' + s + ' chrome')
    elif ua.find(r'iron') != -1:
        browser.append(w + ' ' + s + ' iron')
    elif ua.find(r'applewebkit/') != -1:
        applewebkit_ver_matches = re.search(r'version\/(\d+)', ua)
        if applewebkit_ver_matches is not None:
            browser.append(w + ' ' + s + ' ' + s + applewebkit_ver_matches.group(1))
        else:
            browser.append(w + ' ' + s)
    elif ua.find(r'mozilla/') != -1:
        browser.append(g)
    elif ua.find(r'chrome/') != -1:
        browser.append('chrome')
    elif ua.find(r'wget/') != -1:
        browser.append('wget')
    elif ua.find(r'lynx/') != -1:
        browser.append('lynx')
    elif ua.find(r'curl/') != -1:
        browser.append('curl')
    elif ua.find(r'unknown'):
        browser.append('unknown')

def analyze_platform(ua, browser):
    '''
        Determine the platform from the user agent.
    '''

    if ua.find('j2me') != -1:
        browser.append('j2me')
    elif ua.find('java') != -1:
        browser.append('java')
    elif ua.find('python') != -1:
        browser.append('python')
    elif ua.find('iphone') != -1:
        browser.append('iphone')
    elif ua.find('ipod') != -1:
        browser.append('ipod')
    elif ua.find('ipad') != -1:
        browser.append('ipad')
    elif ua.find('android') != -1:
        browser.append('android')
    elif ua.find('blackberry') != -1:
        browser.append('blackberry')
    elif ua.find('mobile') != -1:
        browser.append('mobile')
    elif ua.find('mac') != -1 or ua.find('darwin') != -1:
        browser.append('mac')
    elif ua.find('webtv') != -1:
        browser.append('webtv')
    elif ua.find('win') != -1:
        browser.append('win')
    elif ua.find('freebsd') != -1:
        browser.append('freebsd')
    elif ua.find('x11') != -1 or ua.find('linux') != -1:
        browser.append('linux')

def browser_types(request):
    ''' Returns list of compatible browser types from Django request. '''

    ua = request.META.get('HTTP_USER_AGENT', 'unknown')
    return user_agent_tags(ua)

def is_primitive_browser(request):
    ''' Returns whether browser is probably primitive.
        A browser is primitive if it can not properly display javascript or css. '''

    b = browser_types(request)
    dumb = (
        'unknown' in b or
        'wget' in b or
        'lynx' in b or
        'curl' in b or
        # presumably a modern browser written in a language will not identify as that language
        'python' in b or
        'java' in b)
    # log('is_primitive_browser: {}'.format(dumb))
    return dumb


def is_known_bot(browser, other):
    ''' Returns True if this access is from a known bot.

        >>> from denova.net.web_log_parser import LogLine
        >>> line = '54.186.50.83 - - [21/Oct/2015:08:26:45 +0000] "GET /server/prices/ HTTP/1.1" 200 57201 "-" "BusinessBot: Nathan@lead-caddy.com"'
        >>> access = LogLine(line)
        >>> is_known_bot(access.browser_name, 'BusinessBot: Nathan@lead-caddy.com')
        True
        >>> is_known_bot(access.browser_name, access.other)
        True
        >>> is_known_bot('curl', 'agent: curl/7.26.0')
        True
        >>> b = '(compatible; Googlebot/2.1; +http://www.google.com/bot.html) agent: Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html'
        >>> is_known_bot('Mozilla', b)
        True
        >>> is_known_bot('ltx71', '(http://ltx71.com/)')
        True
        >>> is_known_bot('', '')
        False
        >>> line = '137.226.113.13 - - [25/Jan/2020:00:45:52 +0000] "GET / HTTP/1.1" 200 145465 "-" "Mozilla/5.0 zgrab/0.x"'
        >>> access = LogLine(line)
        >>> is_known_bot(access.browser_name, access.other)
        True
    '''

    try:
        browser_lc = browser.lower()
        other_lc = other.lower()

        known_bot = (
            ('mozilla' in browser_lc and len(other) <= 0) or # fake mozilla
            browser_lc.startswith('java') or # language lib
            'bot' in browser_lc or           # bot
            'bot' in other_lc or
            'spider' in browser_lc or        # spider
            'spider' in other_lc or
            'baiduspider' in browser_lc or   # BaiduSpider
            'baiduspider' in other_lc or
            'walker' in browser_lc or        # walker
            'walker' in other_lc or
            'crawl' in browser_lc or         # crawl
            'crawl' in other_lc or
            'python-urllib' in browser_lc or
            'libwww-perl' in browser_lc or
            'yahoo' in browser_lc or         # Yahoo
            'slurp' in browser_lc or         # slurp
            'slurp' in other_lc or
            'sleuth' in browser_lc or        # Xenu Link Sleuth
            'sleuth' in other_lc or          # Xenu Link Sleuth
            'curl' in browser_lc or          # curl
            'python' in browser_lc or        # python
            'perl' in browser_lc or          # perl
            'nambu' in browser_lc or         # nambu
            'docomo' in browser_lc or        # DoCoMo
            'digext' in browser_lc or        # DigExt
            'morfeus' in browser_lc or       # Morfeus
            'twitt' in browser_lc or         # twitt
            'sphere' in browser_lc or        # sphere
            'pear' in browser_lc or          # PEAR
            'wordpress' in browser_lc or     # wordpress
            'radian' in browser_lc or        # radian
            'eventbox' in browser_lc or      # eventbox
            'monitor' in browser_lc or       # monitor
            'mechanize' in browser_lc or     # mechanize
            'facebookexternal' in browser_lc or # facebookexternal
            'scoutjet' in other_lc or        # Scoutjet
            'yandex' in browser_lc or        # Yandex
            'yandex' in other_lc or
            'nerdybot' in browser_lc or      # NerdyBot
            'nerdybot' in other_lc or
            'archiver' in browser_lc or      # Archiver
            'archiver' in other_lc or
            'ia_archiver' in browser_lc or   # Alexa
            'ia_archiver' in other_lc or
            'qqdownload' in browser_lc or    # QQ
            'qqdownload' in other_lc or
            'ask jeeves' in browser_lc or    # Ask Jeeves
            'ask jeeves' in other_lc or
            'ltx71' in browser_lc or         # Scan internet for security research purposes
            'ltx71' in other_lc or
            'zgrab' in browser_lc or         # Conformance Testing by a student
            'zgrab' in other_lc or
            'nmap scripting engine' in browser_lc or  # Nmap Scripting Engine
            'nmap scripting engine' in other_lc
            )
    except: # 'bare except' because it catches more than "except Exception"
        known_bot = False
        log(format_exc())

    log(f'"{browser} {other}" known bot: {known_bot}')

    return known_bot

def is_known_harvester(user_agent):
    ''' Returns True if this access is from a known harvester.

        >>> is_known_harvester('Java/1.4.1_04')
        True
        >>> is_known_harvester('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET')
        True
        >>> is_known_harvester('')
        False
    '''

    if user_agent and user_agent.strip():
        known_harvester = (user_agent.startswith('Java/1.4') or
                           user_agent.startswith('Java/1.5') or
                           user_agent.startswith('Java/1.6.') or
                           user_agent.startswith('Java/1.7') or
                           user_agent.startswith('Java/1.8') or
                           user_agent.startswith('Mozilla/4.0 (compatible ; MSIE 6.0; Windows NT 5.1)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows 98)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt; DTS Agent') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1)') or
                           user_agent.startswith('Mozilla/4.0(compatible; MSIE 5.0; Windows 98; DigExt)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)') or
                           user_agent.startswith('MJ12bot/v1.0.8 (http://majestic12.co.uk/bot.php?+)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 5.01; Windows NT 5.0)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; .NET CLR 1.1.4322)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 5.0; Windows NT; DigExt)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0)') or
                           user_agent.startswith('Mozilla/5.0 (compatible; Googlebot/2.1; http://www.google.com/bot.html)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows 98) XX') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322; .NET CLR 2.0.50727)') or
                           user_agent.startswith('Mozilla/5.0 (Windows; U; Windows NT 6.1; en-us; rv:1.9.2.3) Gecko/20100401 YFF35 Firefox/3.6.3') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; en)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; .NET CLR 1.1.4322)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; .NET CLR 1') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET') or
                           user_agent.startswith('ISC Systems iRc Search 2.1') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; .NET CLR 1.1.4322; .NET CLR 2.0.50727; .NET CLR 3.0.04506.30; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729)') or
                           user_agent.startswith('Mozilla/3.0 (compatible)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 2.0.50727)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.1.4322; InfoPath.1)') or
                           user_agent.startswith('Mozilla/3.0 (compatible; Indy Library)') or
                           user_agent.startswith('Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.0.7) Gecko/20060909 Firefox/1.5.0.7') or
                           user_agent.startswith('Missigua Locator 1.9') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET)') or
                           user_agent.startswith('Wells Search II') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; .NET CLR 1.0.3705)') or
                           user_agent.startswith('Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US; rv:1.9.2.11) Gecko/20101012 Firefox/3.6.11 GTB7.1 ( .NET CLR 3.5.30729; .NET4.0E)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; MyIE2; .NET CLR 1.1.4322)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; .NET CLR 1.1.4322)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 5.0; Windows NT)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; .NET CLR 1.0.3705;') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Win32)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0; SLCC1; .NET CLR 2.0.50727; Media Center PC 5.0; .NET CLR 3.0.04506)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; .NET CLR 2.0.50727)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; InfoPath.1)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; .NET CLR 1.1.4322; .NET CLR 2.0.50727)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; InfoPath.2)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0b; Windows NT 6.0)') or
                           user_agent.startswith('Opera/9.0 (Windows NT 5.1; U; en)') or
                           user_agent.startswith('Microsoft URL Control - 6.01.9782') or
                           user_agent.startswith('Microsoft URL Control - 6.00.8862') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Win32)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0; SLCC1; .NET CLR 2.0.50727; .NET CLR 3.0.04506)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE8.0; Windows NT 6.0) .NET CLR 2.0.50727)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; FunWebProducts)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 4.0)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0; SLCC1; .NET CLR 2.0.50727; Media Center PC 5.0; .NET CLR 3.0.04506; InfoPath.2)') or
                           user_agent.startswith('Mozilla/5.0 (Windows NT 5.1; U; en) Opera 8.01') or
                           user_agent.startswith('Opera/9.00 (Windows NT 5.1; U; en)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; Win64; AMD64)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; .NET CLR 1.0.3705; .NET CLR 1.1.4322)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; FREE; .NET CLR 1.1.4322)') or
                           user_agent.startswith('8484 Boston Project v 1.0') or
                           user_agent.startswith('Mozilla/5.0 (compatible; MegaIndex.ru/2.0; +https://www.megaindex.ru/?tab=linkAnalyze)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1; .NET CLR 1.1.4322; .NET CLR 2.0.50727; .NET CLR 3.0.04506.30)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1; MRA 4.3 (build 01218))') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; .NET CLR 1.0.3705)') or
                           user_agent.startswith('Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; SV1; .NET CLR 1.1.4322; .NET CLR 2.0.50727)')
                          )
        log(f'{user_agent} known harvester: {known_harvester}')

    else:
        log.debug('no user agent')
        known_harvester = False

    return known_harvester


def is_known_spammer(referer):
    '''Return True if the referer is a known spammer. '''

    return (
        'semalt.com' in referer or
        'buttons-for-website.com' in referer or
        'buttons-for-your-website.com' in referer or
        'see-your-website-here.com' in referer or
        'makemoneyonline.com' in referer or
        'domainsigma' in referer or
        'darodar.com' in referer or
        'econom.co' in referer or
        'ilovevitaly.co' in referer or
        'best-seo-offer.com' in referer or
        'domini.cat' in referer or
        'savetubevideo.com' in referer or
        'kambasoft.com' in referer or
        'http://123.249.24' in referer
          )

def get_agent_info(agent):
    '''Get the browser name, version, and other data from the agent.

       If agent not defined, return empty strings.
    '''

    if agent:
        User_Agent_Format = re.compile(r"(?P<browser>.*?)/(?P<version>[0-9\.]*)\s*(?P<other>\(?.*\)?)")
        if not isinstance(agent, str):
            agent = agent.decode()
        m = User_Agent_Format.search(agent)
        if m:
            browser_name = m.group('browser')
            browser_version = m.group('version')
            other = m.group('other')
        else:
            browser_name = browser_version = ''
            other = agent
    else:
        browser_name = browser_version = other = ''

    return browser_name, browser_version, other

def get_browser_name(user_agent_tags):
    if user_agent_tags is None:
        browser_name = None
    else:
        browser_name = user_agent_tags[0]

    return browser_name


if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
    Extra HTTP functions.
    http_addons is named to avoid conflict with python's http pacakge.

    Copyright 2013-2020 DeNova
    Last modified: 2020-10-20

    HTTP is a byte oriented protocol.
    In general, this module expects and return bytes.
    Param dicts are an exception. Param names and values are strings.

    As a convenience, in some cases string parameters will be automatically converted to bytes.

    (Implementaion of these type guidelines are a work in progress.)

    To do: Tests, such as for params_to_str() and params_to_bytes()

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import ssl
from http import HTTPStatus
from http.client import HTTPConnection, HTTPSConnection
from http.client import responses as client_responses
from traceback import format_exc
from urllib.parse import urlsplit, urlunsplit

from denova.python.log import get_log
from denova.python.format import to_bytes, to_string
from denova.python.utils import gunzip
from denova.net.openssl import verify_certificate, SELF_SIGNED_CERT_ERR_MSG, EXPIRED_CERT_ERR_MSG
from denova.python.dict import CaseInsensitiveDict

log = get_log()

HTTP_EOL = b'\r\n'
HTTP_SEPARATOR = HTTP_EOL + HTTP_EOL

# dict to look up HTTP status code by name
code = {}
# yes, prospector, this is an iterable
for hs in HTTPStatus:
    code[hs.name] = hs.value

ports = dict(
    http = 80,
    https = 443)

def TEST_get_response(url, proxy=None, cert_file=None):
    ''' Get an HttpResponse for the url '''

    if proxy:
        log.debug(f'get response from "{url}" using proxy "{proxy}"')
        # see python - Using SocksiPy with SSL - Stack Overflow
        #     http://stackoverflow.com/questions/16136916/using-socksipy-with-ssl

        if not cert_file:
            cert_file = '/etc/ssl/certs/ca-certificates.crt'

        proxy_parts = urlsplit(proxy)

        try:
            from socks import socksocket, PROXY_TYPE_SOCKS5
        except:  # 'bare except' because it catches more than "except Exception"
            msg = 'Requires the debian module from python-socksipy'
            log.debug(msg)
            raise Exception(msg)

        s = socksocket()
        s.setproxy(PROXY_TYPE_SOCKS5, proxy_parts.hostname, port=proxy_parts.port)
        s.connect(('check.torproject.org', 443))
        ss = ssl.wrap_socket(s, cert_reqs=ssl.CERT_REQUIRED, ca_certs=cert_file)

        # print "Peer cert: ", ss.getpeercert()

        ss.write(f"""GET / HTTP/1.0\r\nHost: {proxy_parts.hostname}\r\n\r\n""")

        content = []
        while True:
            data = ss.read()
            if not data: break
            content.append(data)

        ss.close()
        result = "".join(content)

    else:
        log.debug(f'get response from {url}')

        url_parts = urlsplit(url)
        if url_parts.scheme == 'https':
            HTTPxConnection = HTTPSConnection
        elif url_parts.scheme == 'http':
            HTTPxConnection = HTTPConnection
            kwargs = {}
        else:
            raise ValueError(f'{url_parts.scheme} not supported')

        conn = HTTPxConnection(url_parts.hostname,
            url_parts.port or ports[url_parts.scheme],
            **kwargs)

        relative_url = urlunsplit(('', '',
                url_parts.path, url_parts.query, url_parts.fragment))
        conn.request('GET', relative_url)

        result = conn.getresponse()

    return result

def get_response(url, proxy=None, cert_file=None):
    '''
        Get an HttpResponse for the url.

        >>> response = get_response('https://denova.com')
        >>> response.status == 200
        True
    '''

    kwargs = {}
    url_parts = urlsplit(url)
    if url_parts.scheme == 'https':
        HTTPxConnection = HTTPSConnection
        if cert_file is not None:
            kwargs = dict(cert_file=cert_file)
    elif url_parts.scheme == 'http':
        HTTPxConnection = HTTPConnection
    else:
        raise ValueError(f'{url_parts.scheme} not supported')

    if proxy is None:
        log.debug(f'get response from {url}')
        connection = HTTPxConnection(url_parts.hostname,
                                     url_parts.port or ports[url_parts.scheme],
                                     **kwargs)

    else:
        log.debug(f'get response from "{url}" using proxy "{proxy}"')
        # weirdly, http.client.HTTPConnection() gets the proxy, and set_tunnel() gets the destination domain
        proxy_parts = urlsplit(proxy)
        connection = HTTPxConnection(
            proxy_parts.hostname,
            proxy_parts.port,
            **kwargs)

        connection.set_tunnel(
            url_parts.hostname,
            url_parts.port or ports[url_parts.scheme])

    relative_url = urlunsplit(('', '',
            url_parts.path, url_parts.query, url_parts.fragment))
    connection.request('GET', relative_url)

    return connection.getresponse()

def check_response(url, why=None, proxy=None):
    '''
        Check that we got a status code of 200 and some data from the url.

        'url" is the internet address you want to go to.
        'why' is why we're checking.
        'proxy' is http proxy.

        Returns response.

        response.data is set to http data.

        >>> response = check_response('https://denova.com')
        >>> response.status == 200
        True
    '''

    def err_msg(err):
        msg = (f'''{err}
        why: {why}
        testing url: {url}''')

        if proxy:
            msg = msg + (f'''
        proxy url: {proxy}
        ''')

        return msg

    if not why:
        why = 'check http response'

    log.debug(f'url: {url}')
    try:
        response = get_response(url, proxy=proxy)
    except:  # 'bare except' because it catches more than "except Exception"
        print(err_msg('error in get_response()'))
        raise

    # check for OK response
    assert response.status == 200, err_msg(f'bad status: {response.status}')

    # check data exists
    response.data = response.read()
    assert len(response.data) > 0, err_msg('no data')

    return response

def parse_request(request):
    '''
        Parse raw http request data into prefix, params, and data.

        'request' must be a string or bytes.

        Returns:
        'prefix' is same type as 'request'.
        'params' is a dict with names and values as strings.
        'data' is same type as 'request', or None.
    '''

    if type(request) is str:
        request = to_bytes(request)
    if type(request) is not bytes:
        raise TypeError("'request' must be bytes")

    prefix, _, remainder = request.partition(HTTP_EOL)

    if HTTP_SEPARATOR in remainder:
        raw_params, _, content = remainder.partition(HTTP_SEPARATOR)
    else:
        raw_params = remainder
        content = None
    params = parse_params(raw_params.split(HTTP_EOL))

    return prefix, params, content

def parse_response(response):
    '''
        Parse raw http response data into (prefix, params, content).

        Returns:
        'prefix' and 'content' are byte sequences.
        'params' is a dict.
    '''

    if type(response) is str:
        response = to_bytes(response)
    if type(response) is not bytes:
        raise TypeError("'response' must be bytes")

    header, _, content = response.partition(HTTP_SEPARATOR)
    prefix, params = parse_header(header)
    params, content = uncompress_content(params, content)

    return prefix, params, content

def unparse_response(prefix, params, data):
    ''' Constructs an http response.

        This is essentially the reverse of parse_response().
    '''
    return (prefix + HTTP_EOL +
            params_to_bytes(params) + HTTP_SEPARATOR +
            to_bytes(data))

def parse_prefix(prefix):
    ''' Parse raw http prefix.

        Returns (http_method, local_url, http_protocol, http_version). '''

    assert isinstance(prefix, bytes), f'type(prefix) must be bytes, not {type(prefix)}'

    try:
        http_method, local_url, protocol_version = prefix.split(b' ')
    except ValueError:
        log(f'ValueError in prefix: "{prefix}"')
        raise

    try:
        http_protocol, http_version = protocol_version.split(b'/')
    except ValueError:
        log(f'ValueError in protocol_version: "{protocol_version}"')
        raise

    return http_method, local_url, http_protocol, http_version

def parse_header(header):
    ''' Parse raw http header into prefix line and a CaseInsensitiveDict. '''

    lines = header.split(HTTP_EOL)
    prefix = lines[0]
    params  = parse_params(lines[1:])

    return prefix, params

def parse_params(lines):
    ''' Parse raw http params.

        Returns a CaseInsensitiveDict. Case of keys is not significant.

        Both names and values are strings. See params_to_bytes() and params_to_str().
    '''

    params = CaseInsensitiveDict()
    for line in lines:
        line = to_string(line).strip()
        if line:
            name, _, value = line.partition(':')
            name = name.strip()
            if name:
                value = value.strip()
                params[name] = value

    return params

def uncompress_content(params, data):
    ''' If content is gzipped, unzip it and set new Content-Length. '''

    if is_text(params) and is_gzipped(params):

        data = gunzip(data)
        del params['Content-Encoding']
        params['Content-Length'] = str(len(data))

    return params, data

def unicode_content(params, data):
    ''' Return content as unicode. If params specify a charset, decode data from that charset. '''

    # see Python Unicode HowTo
    #     http://docs.python.org/2/howto/unicode.html
    #         "Software should only work with Unicode strings internally,
    #         converting to a particular encoding on output."

    if isinstance(data, (bytes, bytearray)):
        charset = content_encoding_charset(params)
        if charset is None:
            # log.debug('no charset in http params')
            unicode_data = to_string(data)
        else:
            try:
                unicode_data = data.decode(charset, 'replace')
            except:  # 'bare except' because it catches more than "except Exception"
                log.debug(f'charset {charset} not supported; trying defaults')
                unicode_data = to_string(data)
            else:
                log.debug(f'decoded http content using charset {charset}')

    elif isinstance(data, str):
        # str is already unicode
        unicode_data = data

    else:
        msg = f'could not decode http content of type {type(data)}'
        log.debug(msg)
        try:
            log.debug(format_exc())
        finally:
            raise ValueError(msg)

    return unicode_data

def create_response(status, params=None, data=None):
    ''' Return raw http response.

        'status' is an integer. http lib defines status constants.

        'params' is an optional dict. 'data' is an optional string. '''

    if params is None:
        params = {}
    if data is None:
        data = b''

    prefix = b'HTTP/1.1 {} {}'.format(status, client_responses[status])
    params['Content-Length'] = len(data)
    response = prefix + HTTP_EOL + params_to_bytes(params) + HTTP_SEPARATOR + to_bytes(data)
    return response

def params_to_str(params):
    ''' Convert params dict to http protocol params string. '''

    lines = []
    for name in params:
        line = f'{camel_case(to_string(name))}: {to_string(params[name])}'
        lines.append(line)

    params_as_str = to_string(HTTP_EOL).join(lines)
    return params_as_str

def params_to_bytes(params):
    ''' Convert params dict to http protocol params byte sequence. '''

    return to_bytes(params_to_str(params))

def header(data):
    ''' Parse header from raw http data '''

    header_data, _, _ = data.partition(HTTP_SEPARATOR)
    return header_data

def is_content_type(params, prefix):
    ''' Return True if Content-Type sarts with prefix, else False. '''

    result = (
        'Content-Type' in params and
        params['Content-Type'].startswith(prefix))
    return result

def is_html(params):
    return is_content_type(params, 'text/html')

def is_text(params):
    ''' Return True if params indicate content is text, else False. '''

    return is_content_type(params, 'text/')

def is_app_data(params):
    ''' Return True if params indicate content is application data, else False. '''

    return is_content_type(params, 'application/')

def is_image(params):
    ''' Return True if params indicate content is image data, else False. '''

    return is_content_type(params, 'image/')

def content_encoding_charset(params):
    ''' Parse content-encoding charset from http response. '''

    charset = None
    if 'Content-Type' in params:
        content_type =  params['Content-Type']
        # content-type = text/html; charset=UTF-8
        for ct_param in content_type.split(';'):
            if charset is None:
                if '=' in ct_param:
                    name, value = ct_param.split('=')
                    name = name.strip()
                    if name == 'charset':
                        charset = value.strip()

    return charset

def is_gzipped(params):
    ''' Return True if params indicate content is gzipped, else False. '''

    return (
        'Content-Encoding' in params and
        'gzip' in params['Content-Encoding'])

def verify_cert_locally(host, port):
    ''' Verify the site's certificate using openssl locally. '''

    log.warning('verify_cert_locally() bypasses any proxy such as Tor, so it may leak DNS info')

    # Verify the cert is ok before proceeding
    log.debug(f'verifying cert for: {host}:{port}')
    ok, original_cert, cert_error_details = verify_certificate(host, port)
    log.debug(f'{host}:{port} cert ok: {ok}')

    if not ok:
        log.debug(cert_error_details)

        # if the cert is self signed or expired, let the user decide what to do
        if SELF_SIGNED_CERT_ERR_MSG in cert_error_details:
            log.debug('cert is self.signed')
            ok = True
        elif EXPIRED_CERT_ERR_MSG in cert_error_details:
            log.debug('cert is expired')
            ok = True

    return ok, original_cert, cert_error_details


def camel_case(name):
    ''' Upper case the first letter of each word in the name. '''

    assert isinstance(name, str)
    return name.replace('-', ' ').title().replace(' ', '-')


if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
    Utility classes and functions.

    Copyright 2009-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import os
import os.path
import re
import sys
from traceback import format_exc
from urllib.parse import urljoin

try:
    from django.conf import settings
    from django.db.models import CharField, EmailField
    from django.utils.encoding import DjangoUnicodeDecodeError
except ModuleNotFoundError:
    sys.exit('Django required')

from denova.python.log import get_log
from denova.python.utils import is_string


log = get_log()

def is_secure_connection(request):
    ''' Check if connection is secure.

        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('https://127.0.0.1/bitcoin/')
        >>> request.META['HTTP_X_SCHEME'] = 'https'
        >>> is_secure_connection(request)
        True

        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('https://127.0.0.1/bitcoin/')
        >>> request.META['wsgi.url_scheme'] = 'https'
        >>> is_secure_connection(request)
        True

        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('http://127.0.0.1/bitcoin/')
        >>> is_secure_connection(request)
        False
    '''

    secure = False
    try:
        if 'HTTP_X_SCHEME' in request.META:
            secure = request.META['HTTP_X_SCHEME'] == 'https'
        elif 'wsgi.url_scheme' in request.META:
            secure = request.META['wsgi.url_scheme'] == 'https'
    except Exception:
        log(format_exc())

    return secure

def get_websocket_url(request):
    '''
        Get the websocket url based on the http connection.

        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('https://127.0.0.1/bitcoin/')
        >>> request.META['HTTP_X_SCHEME'] = 'https'
        >>> get_websocket_url(request)
        'wss://127.0.0.1'

        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('http://127.0.0.1/bitcoin/')
        >>> get_websocket_url(request)
        'ws://127.0.0.1'
    '''

    if 'HTTP_HOST' in request.META:
        host = request.META['HTTP_HOST']
    elif 'REMOTE_ADDR' in request.META:
        host = request.META['REMOTE_ADDR']
    else:
        host = None

    if host:
        if is_secure_connection(request):
            url = f'wss://{host}'
        else:
            url = f'ws://{host}'
    else:
        url = None

    return url

def django_error_page_response(request, error=None):
    ''' Return a response with Django's error page.

        If settings.DEBUG is True, Django automatically shows a useful
        error page for exceptions in views. But sometimes an exception
        isn't propogated out of the view, such as when the exception
        occurs in a separate thread. This shows the Django error page
        for any exception.

        If error is not present or is None, returns an error page for the
        last exception.

        Example:
            error = None
            ...
            # in separate thread
            error = sys.exc_info()
            ...
            # in parent thread
            show_django_error_page(error)
        '''

    from django.views.debug import technical_500_response
    # error should be sys.exc_info() from an earlier except block
    if not error:
        error = sys.exc_info()
    exc_type, exc_value, tb = error
    response = technical_500_response(request, exc_type, exc_value, tb)

    return response


def is_django_error_page(html):
    ''' Returns True if this html contains a Django error page,
        else returns False.'''

    django_error_1 = b"You're seeing this error because you have"
    django_error_2 = b'display a standard 500 page'

    try:
        smart_html = html #smart_text(html)
    except DjangoUnicodeDecodeError:
        # definitely not a django html error page
        result = False
    else:
        result = (django_error_1 in smart_html) and (django_error_2 in smart_html)

    return result

def get_json_dir():
    '''
        Get the directory name where json output from the database are stored.

        >>> dirname = get_json_dir()
        >>> dirname.endswith('data/json')
        True
    '''

    return  os.path.join(settings.DATA_DIR, 'json')

def first_instance(model):
    ''' Return first instance of a model, or None if none.

        When django insists on a default for a ForeignKey field definition, this works.
        Because the function in default=function can't have any params, use something like::

            class MyModel(Model)

                ...

                def other_model_default():
                    return first_instance(MyOtherModel)

        >>> from django.contrib.sites.models import Site
        >>> from django.core.management import call_command
        >>> sites = Site.objects.all()
        >>> if len(sites) > 0:
        ...     for site in sites:
        ...         __ = site.delete()
        >>> CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
        >>> fixture_name = os.path.join(CURRENT_DIR, 'fixtures', 'sites.site.json')
        >>> call_command("loaddata", f"{fixture_name}", verbosity=0)
        >>> first_instance(Site)
        <Site: test.com>
    '''

    objects = None
    default_object = None

    try:
        objects = model.objects.filter(is_active=True)
        log(objects)
    except Exception:
        try:
            objects = model.objects.all()
            for obj in objects.order_by('pk'):
                log(obj.pk)
        except Exception:
            log(format_exc())

    if objects:
        if len(objects) > 0:
            objects = objects.order_by('pk')
            default_object = objects[0]

    return default_object

def get_remote_ip(request):
    '''
        Get the remote ip. If there is a forwarder, assume the first IP
        address (if there are more than 1) is the original machine's address.

        Otherwise, use the remote addr.

        Any errors, return 0.0.0.0

        >>> from django.test import RequestFactory
        >>> factory = RequestFactory()
        >>> request = factory.get('https://127.0.0.1/bitcoin/')
        >>> get_remote_ip(request)
        '127.0.0.1'
    '''

    Unknown_IP = '0.0.0.0'

    if request:
        try:
            # if we're using a reverse proxy, the ip is the proxy's ip address
            remote_addr = request.META.get('REMOTE_ADDR', '')
            forwarder = request.META.get('HTTP_X_FORWARDED_FOR', '')
            if forwarder and forwarder is not None and len(forwarder) > 0:
                m = re.match('(.*?),.*?', forwarder)
                if m:
                    remote_ip = m.group(1)
                else:
                    remote_ip = forwarder
            else:
                remote_ip = remote_addr

            if not remote_ip or remote_ip is None or len(remote_ip) <= 0:
                remote_ip = Unknown_IP
        except Exception:
            log(format_exc())
            remote_ip = Unknown_IP
    else:
        remote_ip = Unknown_IP
        log('no request so returning unknown ip address')

    return remote_ip

def get_absolute_url(url, home_url, request=None):
    '''
        Return an absolute url from a relative url
        adapting for protocol if request included.

        >>> get_absolute_url('bitcoin', 'http://127.0.0.1')
        'http://127.0.0.1/bitcoin'
    '''

    final_home_url = home_url

    if url.startswith('/'):
        url = url[1:]

    try:
        if request is not None and request.META.get('HTTP_REFERER') is not None:
            # use the same protocol for the new url
            referer = request.META.get('HTTP_REFERER')
            if referer.find('://' + settings.TOP_LEVEL_DOMAIN) > 0 and referer.lower().startswith('https'):
                index = final_home_url.find('://')
                if index >= 0:
                    final_home_url = 'https' + final_home_url[index]
                    log(f'final url: {final_home_url}')
    except Exception:
        pass

    return urljoin(final_home_url, url)

def strip_input(data):
    '''Strip the leading and trailing spaces.

        >>> data = "  This is a line without end spaces.  "
        >>> strip_input(data)
        'This is a line without end spaces.'
    '''

    try:
        if data is not None:
            if is_string(data) or isinstance(data, CharField):
                data = data.strip()

            elif isinstance(data, EmailField):
                data = f'{data}'
                data = data.strip()

            elif isinstance(data, bytes):
                data = data.decode().strip()
    except Exception:
        log(format_exc())

    return data


if __name__ == "__main__":
    import doctest
    doctest.testmod()
